
/* (c) 1995 VictorMaxx Technologies  */
/* Modified 12/95 by Haim Bodek      */

/*********************************************************************/
/**       CTM.C       CyberMaxx Tracking Module routines            **/
/**                   Serial Communications routines                **/
/**                   Requires CTM.H                                **/
/*********************************************************************/

/* Bibliography: */
/* Serial Communications: A C++ Deveoper's Guide. Nelson. */
/* The Black Art of 3D Game Programming. LaMothe. 1995.   */

#include <dos.h>

/* GLOBAL VARIABLES FOR INTERRUPT SERVICE ROUTINE */

void (_interrupt _far *Old_Serial_ISR)();  // holds old interrupt handler
char serial_buffer[BUFFSIZE];          // the receive buffer
int serial_end   = -1;                 // indexes into receive buffer
int serial_start = -1;
int serial_ch;
int char_ready=0;                      // current character and ready flag
int old_int_mask;                      // the old interrupt mask on the PIC
int open_port;                         // the currently open port

/* OLD SETTING GLOBAL VARIABLE */

int obaud, obits, oparity, ostopbits;

/*********************************************************************/
/**  int port_exist (int comport)                                   **/
/**  Determines if Com Port 1 or 2 physically exists                **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: TRUE or FALSE                                          **/
/*********************************************************************/

int port_exist(int comport)
   {
   int temp;
   if (!((comport==1) || (comport==2))) return FALSE;
   outp(UARTS[comport]+FCR,0);
   temp=inp(UARTS[comport]+IIR);
   if ((temp&0xf8)!=0) return FALSE;
   else return TRUE;
   }

/*********************************************************************/
/**  int get_settings(int comport, int *oldbaud, int *oldbits,      **/
/**                   int *oldparity, int *oldstopbits)             **/
/**  Stores the current settings of a given Com Port                **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: TRUE or FALSE depending on success                     **/
/**  int *oldbaud: the previous baud rate 9600, 19200, etc.         **/
/**  int *bits: the previous bits 5,6,7,8                           **/
/**  int *oldparity: previous parity NONE, EVEN, ODD, MARK, SPACE   **/
/**  int *oldstopbits: previous stop bits 1,2                       **/
/*********************************************************************/


int get_settings(int comport, int *oldbaud, int *oldbits, int *oldparity, 
                  int *oldstopbits)
  {
  int temp;
  int lcr;
  int divisor_low;
  int divisor_high;
  
  if (!((comport==1) || (comport==2))) return FALSE;

  lcr=inp(UARTS[comport]+LCR);
  lcr|=LCR_DLAB;
  
  _disable();
  outp(UARTS[comport]+LCR,lcr);
  divisor_low=inp(UARTS[comport]+DLL);
  divisor_high=inp(UARTS[comport]+DLH);
  lcr&=~LCR_DLAB;
  outp(UARTS[comport]+LCR,lcr);
  _enable();

  /* determine baud rate */
  if (divisor_high | divisor_low)
    *oldbaud=(115200L / ((divisor_high << 8) + divisor_low));
  else return FALSE;
 
  /* determine bits */
  temp=lcr & LCR_WORD_LENGTH_MASK;
  if (temp==0) *oldbits=5;
  else if (temp==LCR_WORD_0) *oldbits=6;
  else if (temp==LCR_WORD_1) *oldbits=7;
  else if (temp==LCR_WORD_0 + LCR_WORD_1) *oldbits=8;
  else return FALSE;
  
  /* determine parity */
  temp=lcr&LCR_PARITY_MASK;
  if (temp==LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT+LCR_STICK_PARITY)
    *oldparity=SPACE;
  else if (temp==LCR_PARITY_ENABLE+LCR_STICK_PARITY) *oldparity=MARK;
  else if (temp==LCR_PARITY_ENABLE) *oldparity=ODD;
  else if (temp==(LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT)) *oldparity=EVEN;
  else *oldparity=NONE;
  
  /* determine stop bits */
  temp=lcr & LCR_STOP_BITS;
  if (temp==0) *oldstopbits=1;
  else if (temp==1) *oldstopbits=2;
  else return FALSE;
  
  return TRUE;
  }

/*********************************************************************/
/**  int set_port (int comport, int baud, int bits,                 **/
/**                   int parity, int stopbits)                     **/
/**  Changes the current settings of a given Com Port               **/
/**  Input: comport: 1 or 2                                         **/
/**         baud: 9600, 19200, etc.                                 **/
/**         bits: 5,6,7,8                                           **/
/**         parity: NONE, ODD, EVEN, MARK, SPACE                    **/
/**         stopbits: 1,2
/**  Output: TRUE or FALSE depending on success                     **/
/*********************************************************************/

int set_port(int comport, int baud, int bits, int parity, int stopbits)
  
  {
  int lcr;
  int divisor_high;
  int divisor_low;
  long result_baud;
  
  if (!((comport==1) || (comport==2))) return FALSE;
  
  divisor_low = (int) ((115200L / baud) & 0xff);
  divisor_high= (int) ((115200L / baud) >> 8);
  result_baud=115200L/(115200L/baud);
  if (result_baud!=baud) return FALSE;
  lcr = inp(UARTS[comport] +LCR);
  lcr|=LCR_DLAB;
  
  _disable();
  outp(UARTS[comport]+LCR, lcr);
  outp(UARTS[comport]+DLL, divisor_low);
  outp(UARTS[comport]+DLH, divisor_high);
  lcr &=~LCR_DLAB;
  outp(UARTS[comport]+LCR, lcr);
  _enable();

  lcr &=~LCR_PARITY_MASK;
  if (parity==NONE);
  else if (parity==ODD) lcr|=LCR_PARITY_ENABLE;
  else if (parity==EVEN) lcr|=LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT;
  else if (parity==MARK) lcr|=LCR_PARITY_ENABLE+LCR_STICK_PARITY;
  else if (parity==SPACE) lcr|=LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT+
        LCR_STICK_PARITY;
  else return FALSE;

  lcr&=~LCR_WORD_LENGTH_MASK;
  if (bits==5);
  else if (bits==6) lcr|=LCR_WORD_0;
  else if (bits==7) lcr|=LCR_WORD_1;
  else if (bits==8) lcr|=LCR_WORD_0+LCR_WORD_1;
  else return FALSE;

  lcr&=~LCR_STOP_BITS;
  if (stopbits==1);
  else if (stopbits==2) lcr|=LCR_STOP_BITS;
  else return FALSE;
  
  outp(UARTS[comport]+LCR, lcr);
  return TRUE;
  }

/*********************************************************************/
/** void _interrupt _far Serial_ISR(void)                           **/
/** Interrupt Service Routine:                                      **/
/**     Used Internally by startISR(int) and closeISR(int)          **/
/** If you are using protected mode, you will want to install your  **/
/** own bimodal interrupt handler in place of this routine.         **/
/** The next update of this library will have protected mode support**/
/*********************************************************************/

void _interrupt _far Serial_ISR(void)
  {
  _enable();
  serial_ch=inp(UARTS[open_port]);
  if (++serial_end > BUFFSIZE-1) serial_end=0;
  serial_buffer[serial_end]=serial_ch;
  ++char_ready;
  outp(PIC_ICR,0x20);
  }

/*********************************************************************/
/**  void startISR(int comport)                                     **/
/**  Initiates the Interrupt Service routine for a give Com Port    **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: none                                                   **/
/*********************************************************************/


void startISR(int comport)
  {
  unsigned char data;
  open_port=comport;
  data=inp(UARTS[comport]+MCR);
  data|=SERIAL_GP02;
  outp(UARTS[comport]+MCR,data);
  outp(UARTS[comport]+IER,1);
  
  if (comport==1)
    {
    Old_Serial_ISR = _dos_getvect(INT_SERIAL_PORT_0);
    _dos_setvect(INT_SERIAL_PORT_0,Serial_ISR);
    }
  else
    {
    Old_Serial_ISR = _dos_getvect(INT_SERIAL_PORT_1);
    _dos_setvect(INT_SERIAL_PORT_1,Serial_ISR);
    }
  
  old_int_mask=inp(PIC_IMR);
  outp(PIC_IMR,(comport==1) ? (old_int_mask & 0xEF) : (old_int_mask &0xF7));
  }

/*********************************************************************/
/**  void closeISR (int comport)                                    **/
/**  Shuts down the Interrupt Service routine for a give Com Port   **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: none                                                   **/
/*********************************************************************/

void closeISR (int comport)
  {
  unsigned char data;

  data=inp(UARTS[open_port]+MCR);
  data=data & ~SERIAL_GP02;
  outp(UARTS[comport]+MCR,data);

  outp(UARTS[comport] +IER,0);
  outp(PIC_IMR, old_int_mask);

  if (comport==1) _dos_setvect(INT_SERIAL_PORT_0, Old_Serial_ISR);
  else _dos_setvect(INT_SERIAL_PORT_1, Old_Serial_ISR); 
  }

/*********************************************************************/
/**  int read_byte (unsigned int * ch)                              **/
/**  Reads a byte from the input buffer                             **/
/**  Input: none                                                    **/
/**  Output: TRUE or FALSE depending on if a byte was ready         **/
/**  unsigned int *ch: returns read byte                            **/
/*********************************************************************/

int read_byte (unsigned int * ch)
  {
  if (!char_ready) return FALSE; 
  if (serial_end !=serial_start)
    {
    if (++serial_start>BUFFSIZE-1) serial_start=0;
    --char_ready;
    *ch=serial_buffer[serial_start];
    }
  return TRUE;
  }

/*********************************************************************/
/**  int write_byte (int comport, unsigned int CTMByte)             **/
/**  writes a byte to serial port                                   **/
/**  Input: comport 1 or 2, unsigned int CTMByte: character to send **/
/**  Output: TRUE (for right now!)                                  **/
/*********************************************************************/

int write_byte (int comport, unsigned int CTMByte)
  {
  while (!inp(UARTS[comport]+LSR) & 0x20) {}
  _disable();
  outp(UARTS[comport], CTMByte);
  _enable();
  return TRUE;
  }

/*********************************************************************/
/**  int handshakeCTM (int comport)                                 **/
/**  attempts a CTM handhsake                                       **/
/**  Input: comport 1 or 2                                          **/
/**  Output: TRUE or FALSE depending if CTM was deteced             **/
/**  Note: Interrupts may interfere with this routine               **/
/*********************************************************************/

int handshakeCTM (int comport)
  {
  unsigned int ch; long count=0; long wait=0;
  unsigned int a=' ',b=' ',c=' ';
  write_byte (comport, 'X');
  for (;;)
    {
    wait++;
    if (inp(UARTS[comport]+LSR) & 1)
      {
      ch=inp(UARTS[comport]);
      wait=0; count++;
      a=b;
      b=c;
      c=ch;
      if ((a=='C') && ((b=='T') && (c=='M')))
        return TRUE;
      }
    if ((wait>30000) || (count>300)) break;
    }
  return FALSE;
  }    

/*********************************************************************/
/**  void purge_in()                                                **/
/**  purges input buffer                                            **/
/**  Input: none                                                    **/
/**  Output: none                                                   **/
/*********************************************************************/


void purge_in()
  {
  unsigned int ch; int j;
  delay (5);
  for (;;)
    {
    j=0;
    while (!read_byte(&ch))
      {
      delay (1);
      if (j++>5) return;
      }
    }
  }

/*********************************************************************/
/**  int readCTM(int comport, float *yaw, float *pitch, float *roll)**/
/**  read a yaw, pitch, and roll from the CTM                       **/
/**  Input: comport 1 or 2                                          **/
/**  Output: returns TRUE or FALSE depending on sucess of read      **/
/**  float *yaw: returns yaw absolute coordinates                   **/
/**  float *pitch: returns pitch absolute coordinates               **/
/**  float *roll: returns roll absolute coordinates                 **/
/*********************************************************************/


int readCTM(int comport, float *yaw, float *pitch, float *roll)
  {
        unsigned int line[7];
        unsigned int j, i;
        unsigned int csum;
        
        write_byte (comport, 'S');

        for(i=0; i<7; i++) 
          {
          j=0;
          while (!read_byte(&line[i]))
            {
            delay(1);
            if (j++>30) 
              {
              purge_in();
              return FALSE;
              }
            }
          }

        j = (line[0]<<8) + line[1];
        *yaw = j*1.0986328e-2;
        csum=j;

        j = (line[2]<<8) + line[3];
        *pitch = j*5.493164e-3-90.0;
        csum+=j;

        j = (line[4]<<8) + line[5];
        *roll = j*5.493164e-3-90.0;
        csum+=j;

        csum = csum & 0xff;
        
        if(!(*yaw<=360 && *yaw>=0 && *roll>=-48 
                  && *roll<=48 && *pitch>=-48 && *pitch<=48))
                {
                purge_in();
                return FALSE;
                }

        i = 0xff & line[6];
        if ((i - csum)==0) return TRUE; 
        else 
          {
          purge_in();
          return FALSE; 
          }
}

/*********************************************************************/
/**  int detectCTM(int *port)                                       **/
/**  attempts to detect a CTM on Com Port 1 or 2                    **/
/**  Input: none                                                    **/
/**  Output: TRUE or FALSE depending on detect success              **/
/**  int *port: 1 or 2 for detected Com Port                        **/
/*********************************************************************/

int detectCTM(int *port)
  {
  int oldbaud, oldbits, oldparity, oldstopbits;
  int found; int i;
  found=FALSE;

  for (i=1;i<=2;i++)
    {  
    if (port_exist(i)) 
      {
      get_settings(i, &oldbaud, &oldbits, &oldparity, 
                  &oldstopbits);
      set_port (i, 9600, 8, NONE, 1);
      if (handshakeCTM(i)) 
        {
        *port=i;
        found=TRUE;
        }
      set_port (i, oldbaud, oldbits, oldparity, oldstopbits);
      }
    if (found) break;
    }
  return found; 
  }

/*********************************************************************/
/**  int initCTM(int comport)                                       **/
/**  resets the CTM, turns on checksum, starts Interrupt Service    **/                                        
/**  Input: comport 1 or 2                                          **/
/**  Output: TRUE or FALSe depending on success                     **/
/*********************************************************************/

int initCTM(int comport)
  {
  if (!port_exist(comport)) return FALSE;
  if (!get_settings(comport,&obaud,&obits,&oparity,&ostopbits)) return FALSE; //uses globals
  if (!set_port(comport,9600,8,NONE,1)) return FALSE;
  write_byte (comport, 'R');
  delay(100);
  write_byte(comport, 'M'); /* turns on checksum */
  delay (50);
  if (!handshakeCTM(comport)) return FALSE;
  startISR (comport);
  delay (25);
  return TRUE;
  }

/*********************************************************************/
/**  void setCTM(int comport, unsigned int ch)                      **/
/**  Sends a command to the CTM                                     **/
/**  Input: comport 1 or 2                                          **/
/**  command to send (a char)                                       **/
/**  current commands supported by this library are:                **/
/**        ctmAvgFilterOn or 'T'                                    **/
/**        ctmAvgFilterOff or 'W'                                   **/
/**        ctmHysFilterOn or 'P'                                    **/
/**        ctmHysFilterOff or 'O'                                   **/
/**  Output: none                                                   **/
/*********************************************************************/

void setCTM(int comport, unsigned int ch)
  {
  write_byte (comport, ch);
  delay (50);
  }

/*********************************************************************/
/**  int versionCTM(int comport, int *ver, int *revi, unsigned int *subrevi) **/
/**  returnd the firmware version of the CTM                        **/
/**  Input: comport 1 or 2                                          **/
/**  Output: TRUE or FALSE depending on success                     **/
/**  int *ver: version number                                       **/
/**  int *revi: revision number                                     **/
/**  unsigned int *subrevi: surevision letter (char)                **/
/*********************************************************************/

int versionCTM(int comport, int *ver, int *revi, unsigned int *subrevi)
{
  unsigned int in[5];
  int i,j, notit = 1;

  delay(50);
  purge_in();
  write_byte (comport, 'V'); /* request version number */
  delay(50);
        
  for(i=0; i<5; i++) 
    {
    j=0;
    while (!read_byte(&in[i]))
      {
      delay(1);
      if (j++>30) 
        {
        purge_in();
        return FALSE;
        }
      }
    }

  *ver = in[0]-48;
  *revi = ((in[2]-48)*10)+(in[3]-48);
  *subrevi = in[4];
   
  return TRUE;
}

/*********************************************************************/
/**  void closeCTM (int comport)                                    **/
/**  shuts down interrupt service routine, sets old port setting    **/
/**  Input: comport 1 or 2                                          **/
/**  Output: none                                                   **/
/*********************************************************************/

void closeCTM (int comport)
  {
  closeISR(comport);
  set_port(comport,obaud,obits,oparity,ostopbits);
  }
