/* Simple 3D graphics routines */
/* Written by Bernie Roehl, November 1994 for VictorMaxx Technologies */
/* Modified 12/95 by Haim Bodek */
/* Redistribute at will, but charge nothing */


/* Constants needed to set 320x400 Video Mode */

static const unsigned int CRTC_INDEX= 0x3d4; //CRT Controller Index register
static const unsigned int SC_INDEX=   0x3c4; //Sequence Controller Index register
static const unsigned int GC_INDEX=   0x3ce; //Graphics Controller Index register

static short putX, putY, putCOL;

typedef struct { float x, y, z; } Vector;

typedef float Matrix[3][3];

typedef struct
	{
	int start;    /* index into vertex array of start of edge */
	int end;      /* index into vertex array of end of edge */
	int color;    /* color to draw this edge in */
	} Edge;

typedef struct
	{
	Vector location;      /* location of object in the world */
	Matrix orientation;   /* orientation matrix for the object */
	int nvertices;        /* number of vertices in the object */
	Vector *vertices;     /* array of vertices */
	int nedges;           /* number of edges in the object */
	Edge *edges;          /* array of edges */
	} Object;


/* Math routines */

void BuildTrigTables(void);  /* called once at start of application */
float sine(int angle);
float cosine(int angle);

/* Vector operations */

void VectorAdd(Vector *result, Vector *a, Vector *b);
void VectorSubtract(Vector *result, Vector *a, Vector *b);
void VectorRotate(Vector *result, Matrix rotmat, Vector *v);

/* Matrix operations */

void MatrixIdentity(Matrix result);
void MatrixMultiply(Matrix result, Matrix a, Matrix b);
void MatrixMakeXrot(Matrix result, int angle);
void MatrixMakeYrot(Matrix result, int angle);
void MatrixMakeZrot(Matrix result, int angle);

/* 320x400 VGA Display operations */
void Set320x400Mode(void);
int DisplayInitialize(void);
void DisplayUninitialize(void);
void DisplayClear(void);
void DisplayLine(int x1, int y1, int x2, int y2, int color);
void DisplaySetEye(int eye);  /* eye is zero for left, non-zero for right */
extern int DisplayWidth, DisplayHeight;

/* Renderer */

void Render(Object *object, Matrix eye_matrix, float eye_offset, int x_shift);
void RenderSetZoom(float value);

/* inline assembly video routines */

// DisplayClear:
// enable writes to all planes, so we can clear 4 pixels at a time

#pragma aux DisplayClear =    \
	"mov dx, SC_INDEX"       \
	"mov ax, 0xf02"       \        
	"out dx, ax"          \
	"mov ax, 0x0a000"     \
	"mov es, ax"          \
	"sub di, di"          \ 
	"mov ax, di"          \
	"mov cx, 0x8000"      \
	"cld"                 \
	"rep stosw";              

// Set320x400Mode Synopsis:
// First, go to normal 320x200 256-color mode, which is really a
// 320x400 256-color mode with each line scanned twice.
// AH = 0 means mode set, AL = 13h selects
// int is BIOS video interrupt
// Change CPU addressing of video memory to linear (not odd/even,
// chain, or chain 4), to allow us to access all 256K of display
// memory. When this is done, VGA memory will look just like memory
// in modes 10h and 12h, except that each byte of display memory will
// control one 256-color pixel, with 4 adjacent pixels at any given
// address, one pixel per plane.
// Now clear the whole screen, since the mode 13h mode set only
// cleared 64K out of the 256K of display memory. Do this before
// we switch the CRTC out of mode 13h, so we don't see garbage
// on the screen when we make the switch.
// Tweak the mode to 320x400 256-color mode by not scanning each
// line twice.
// Change CRTC scanning from doubleword mode to byte mode, allowing
// the CRTC to scan more than 64K of video data.

#pragma aux Set320x400Mode =               \
	"mov ax, 0x0013"                   \
	"int 0x10"                         \
	"mov dx, SC_INDEX"                 \
	"mov al, 4"                        \
	"out dx, al"                       \
	"inc dx"                           \ 
	"in al, dx"                        \
	"and al, 0xf7"                     \   
	"or al, 0x04"                      \
	"out dx, al"                       \
	"mov dx, GC_INDEX"                 \
	"mov al, 5"                        \
	"out dx, al"                       \
	"inc dx"                           \
	"in al, dx"                        \
	"and al, 0xef"                     \
	"out dx, al"                       \
	"dec dx"                           \
	"mov al, 6"                        \
	"out dx, al"                       \
	"inc dx"                           \
	"in al, dx"                        \
	"and al, 0xfd"                     \
	"out dx, al"                       \
	"mov dx, SC_INDEX"                 \
	"mov ax, 0xf2"                     \  
	"out dx, ax"                       \
	"mov ax, 0x0a000"              \
	"mov es, ax"                       \
	"sub di, di"                       \
	"mov ax, di"                       \
	"mov cx, 0x8000"                   \
	"cld"                              \
	"rep stosw"                        \                    
	"mov dx, CRTC_INDEX"               \
	"mov al, 9"                        \
	"out dx, al"                       \
	"inc dx"                           \
	"in al, dx"                        \
	"and al, 0xe0"                     \
	"out dx, al"                       \
	"dec dx"                           \
	"mov al, 0x14"                     \
	"out dx, al"                       \
	"inc dx"                           \
	"in al, dx"                        \
	"and al, 0xbf"                     \
	"out dx, al"                       \
	"dec dx"                           \
	"mov al, 0x17"                     \
	"out dx, al"                       \
	"inc dx"                           \
	"in al, dx"                        \
	"or al, 0x40"                      \
	"out dx, al";

// putp: Draws a pixel in the specified color at the specified
// location in 320x400 256-color mode.
// Input:
// CX = X coordinate of pixel
// DX = Y coordinate of pixel
// BL = pixel color
// Output: none
// Registers altered: AX, CX, DX, DI, ES
// putp: points to display memory where each 320-pixel row is 80 bytes wide
// since there are 4 pixels at each address divide the X coordinate by 4

#pragma aux putp =                         \
		"mov cx,putX"              \
		"mov dx,putY"              \
		"mov bx,putCOL"            \
		"mov ax,0x0a000"           \
		"mov es,ax"                \    
		"mov ax,80"                \    
		"mul dx"                   \   
		"push cx"                  \    
		"shr cx,1"                 \   
		"shr cx,1"                 \   
		"add ax,cx"                \    
		"mov di,ax"                \
		"pop cx"                   \    
		"and cl,3"                 \   
		"mov ah,1"                 \
		"shl ah,cl"                \   
		"mov al,2"                 \
		"mov dx,SC_INDEX"             \     
		"out dx,ax"                \     
		"mov es:[di],bl";          

/* End of 3d.h */                         


