/* Simple 3D graphics routines */

/* Written by Bernie Roehl, November 1994 for VictorMaxx Technologies */

/* Redistribute at will, but charge nothing and give credit where due */

/* Important note: these routines are *far* from efficient!  They are only
   intended to illustrate some of the basic algorithms for 3D graphics;
   the emphasis is on simple, clear code rather than performance.
   For a much more efficient 3D graphics library, check out AVRIL on
   sunee.uwaterloo.ca (in the pub/avril directory).
 */

/* Assumptions: wireframe, left-handed coordinate system, column vectors,
   matrices multiply vectors as V2 = M * V1.  Distances are in millimeters,
   angles are in tenths of a degree and are stored as integers.
 */

#include "3d.h"

#define MAXVERTICES 100  /* maximum number of vertices allowed in one object */

static Vector eye_coords[MAXVERTICES];  /* stores transformed vertices */

static float zoom = 1.00;  /* zoom factor */

float aspect = 1;
//float aspect = 60.0/54.0;  /* aspect ratio is ratio of the two FOVs */

void Render(Object *object, Matrix eye_matrix, float eye_offset, int x_shift)
	{
	int i;

	for (i = 0; i < object->nvertices; ++i)
		{
		Vector world_coords;

		/* rotate and translate the coordinates into world space */
		VectorRotate(&world_coords, object->orientation, &object->vertices[i]);
		VectorAdd(&world_coords, &world_coords, &object->location);

		/* translate the world coordinates so that the eye is at the
			center of the universe, then rotate them so that view is
			into the positive Z direction */
		/* since the tracker is already at the center of the universe in
			this application, there's no need to do any translation;
			however, we must still do the rotation */
		VectorRotate(&eye_coords[i], eye_matrix, &world_coords);

		/* note that there's no Z-clipping; any object that's partially
			(or entirely) behind the user is simply not rendered */
		if (eye_coords[i].z < 1.00)  /* see if object is too close */
			return;                  /* if so, ignore it */

		/* shift over by half the inter-pupilary distance: */
		eye_coords[i].x += eye_offset;

		/* compute perspective */
		eye_coords[i].x /= eye_coords[i].z;
		eye_coords[i].y /= eye_coords[i].z;

		/* scale to screen space, and adjust for the aspect ratio */
		/* the height is negated, since in virtual space Y increases
			upward, but on the screen it increases downward */
		eye_coords[i].x *= zoom * DisplayWidth;
		eye_coords[i].y *= -zoom * DisplayHeight * aspect;

		/* shift the origin to the center of the screen */
		eye_coords[i].x += DisplayWidth / 2;
		eye_coords[i].y += (DisplayHeight / 2);

		/* add the horizontal shift */
		eye_coords[i].x += x_shift;
		}

	/* Now actually draw the edges of the object */
	for (i = 0; i < object->nedges; ++i)
		{
		Edge *edge = &object->edges[i];
		DisplayLine(eye_coords[edge->start].x, eye_coords[edge->start].y,
			eye_coords[edge->end].x, eye_coords[edge->end].y, edge->color);
		}
	}

void RenderSetZoom(float value)
	{
	zoom = value;
	}

